; Finding free and bound variables, Lambda Calculus chapter
; By Nils M Holm, 2016
;
; Provided under the Creative Common Zero (CC0) license
; (https://creativecommons.org/publicdomain/zero/1.0/)
;
; From the book "Compiling Lambda Calculus": http://t3x.org/clc/

; Examples: (fv '(lambda (x y) (F (lambda z) M)))
;           (bv '(lambda (x y) (F (lambda z) M)))

(define (flatten x)
  (letrec
    ((f (lambda (x r)
          (cond ((null? x)
                  r)
                ((pair? x)
                  (f (car x)
                     (f (cdr x) r)))
                (else
                  (cons x r))))))
    (f x '())))

(define (union a b)
  (cond ((null? a)
          b)
        ((memq (car a) b)
          (union (cdr a) b))
        (else
          (union (cdr a) (cons (car a) b)))))

(define (fv x)
  (let scan ((x x)
             (e '()))
    (cond ((memq x e)
            '())
          ((symbol? x)
            (list x))
          ((not (pair? x))
            '())
          ((eq? 'quote (car x))
            '())
          ((eq? 'if (car x))
            (scan (cdr x) e))
          ((eq? 'lambda (car x))
            (scan (cddr x)
                  (append (flatten (cadr x)) e)))
          (else
            (union (scan (car x) e)
                   (scan (cdr x) e))))))

(define (bv x)
  (let scan ((x x))
    (cond ((not (pair? x))
            '())
          ((eq? 'quote (car x))
            '())
          ((eq? 'if (car x))
            (scan (cdr x)))
          ((eq? 'lambda (car x))
            (union (flatten (cadr x))
                   (scan (cddr x))))
          (else
            (union (scan (car x))
                   (scan (cdr x)))))))
