/*
 * LC-language -> C compiler, runtime, with environment pruning
 * By Nils M Holm, 2016, 2017, 2018
 *
 * Provided under the Creative Common Zero (CC0) license
 * (https://creativecommons.org/publicdomain/zero/1.0/)
 *
 * From the book COMPILING LAMBDA CALCULUS, http://t3x.org/clc/
 */

#include "s9core.h"
#include "s9import.h"

#define	SEG	1000

#define box(x)	cons((x),NIL)
#define unbox	car

cell	E, L, S, T, *Ve, *Vf;
int	K, P, Z;
char	*I;

cell	*R[] = { &E, &L, &S, &T, NULL };

void err(char *s) {
	fprintf(stderr, "L error: %s\n", s);
	exit(1);
}

void init(void) {
	s9_init(R);
	P = 0;
	Z = SEG;
	S = make_vector(Z);
}

cell readexp(void) {
	#define B 256
	char	b[B];
	int	i;
	cell	n, a, p;

	while (isspace(*I))
		I++;
	if ('(' == *I) {
		I++;
		a = NIL;
		n = readexp();
		while (n != VOID) {
			if (NIL == a) {
				save(a = cons(NIL, NIL));
			}
			else {
				p = cons(NIL, NIL);
				cdr(a) = p;
				a = cdr(a);
			}
			car(a) = n;
			n = readexp();
		}
		if (NIL == a)
			return NIL;
		return unsave(1);
	}
	else if (')' == *I) {
		I++;
		return VOID;
	}
	else {
		for (i = 0; !isspace(*I) && ')' != *I; i++) {
			b[i] = *I++;
			if (i >= B) err("symbol too long");
		}
		b[i] = 0;
		return symbol_ref(b);
	}
}

void px(cell x) {
	if (function_p(x)) {
		printf("#<function %d>", (int) cadr(x));
	}
	else if (symbol_p(x)) {
		printf("%s", symbol_name(x));
	}
	else if (NIL == x) {
		printf("()");
	}
	else {
		putchar('(');
		while (!atom_p(x)) {
			px(car(x));
			x = cdr(x);
			if (pair_p(x))
				putchar(' ');
		}
		if (x != NIL) {
			printf(" . ");
			px(x);
		}
		putchar(')');
	}
}

void printexp(cell x) {
	px(x);
	putchar('\n');
}

cell mklitpool(char *s) {
	cell	n, v, *vv;
	int	i;

	I = s;
	n = readexp();
	save(n);
	v = make_vector(length(n));
	unsave(1);
	vv = vector(v);
	for (i = 0; n != NIL; n = cdr(n))
		vv[i++] = car(n);
	return v;
}

void push(cell x) {
	cell	v;

	if (P >= Z) {
		T = x;
		v = make_vector(Z+SEG);
		memcpy(vector(v), vector(S), Z*sizeof(cell));
		Z += SEG;
		S = v;
	}
	vector(S)[P++] = x;
}

cell pushfun(int k, int n) {
	cell	fn, ne;

	ne = make_vector(n+1);
	fn = new_atom(k, ne);
	push(new_atom(T_FUNCTION, fn));
	return ne;
}

int apply(int k, int n) {
	T = vector(S)[P-1];
	P--;
	if (!function_p(T))
		err("application of non-function");
	if (n) {
		vector(E)[n-1] = vector(S)[P-2];
		vector(S)[P-5] = vector(S)[P-1];
		P-=2;
	}
	else {
		push(E);
		push(new_atom(k, NIL));
	}
	E = cddr(T);
	return cadr(T);
}

int ret(int n) {
	int	k;

	T = vector(S)[P-1];
	vector(E)[n] = vector(S)[P-2];
	k = car(vector(S)[P-3]);
	E = vector(S)[P-4];
	P -= 4;
	vector(S)[P-1] = T;
	return k;
}

#define P_cons		cons
#define P_car		car
#define P_cdr		cdr
#define P_pairp(x)	(pair_p(x)? vector(L)[1]: NIL)
#define P_nullp(x)	(NIL == x? vector(L)[1]: NIL)
