# Scheme_3 compiler test suite
# By Nils M Holm, 2016
#
# Provided under the Creative Common Zero (CC0) license
# (https://creativecommons.org/publicdomain/zero/1.0/)
#
# Based on COMPILING LAMBDA CALCULUS, http://t3x.org/clc/
#
# Usage: sh test8.sh

err=0

test() {
	echo -n .
	s9 -l scm8.scm -r "(comp $1)" >_.c
	cc -I /u/include -I . _.c /u/lib/s9core.a
	r=`./a.out`
	if [ "x$r" != "x$2" ]; then
		echo
		echo -n "failed (8): `echo $1 | sed -e 's/  */ /g'`: "
		echo "expected $2, got $r"
		err=1
	fi
}

test "''x" "x"
test "'((lambda (x) x) 'a)" "a"
test "'(((lambda (x) (lambda (y) y)) 'a) 'b)" "b"
test "'(((lambda (x) (lambda (y) x)) 'a) 'b)" "a"
test "'(if 't 'a 'b)" "a"
test "'(if '() 'a 'b)" "b"
test "'(cons 'a 'b)" "(a . b)"
test "'(cons 'a '())" "(a)"
test "'(cons 'a '(b))" "(a b)"
test "'(car '(a b))" "a"
test "'(cdr '(a b))" "(b)"
test "'(pair? '(a . b))" "t"
test "'(pair? '(a))" "t"
test "'(pair? 'a)" "()"
test "'(null? '(a . b))" "()"
test "'(null? '(a))" "()"
test "'(null? 'a)" "()"
test "'(null? '())" "t"
test "'(((lambda (x) (lambda (y) (cons x y))) 'a) 'b)" "(a . b)"
test "'(begin 'a)" "a"
test "'(begin 'a 'b 'c)" "c"
test "'((lambda (x) (set! x 'b)) 'a)" "b"
test "'((lambda (x) ((lambda (x) (set! x 'c)) 'b) x) 'a)" "a"
test "'((lambda (x) ((lambda (y) (set! x 'c)) 'b) x) 'a)" "c"
test "'(((lambda (g) \
          ((lambda (f) (f f)) \
           (lambda (f) (g (lambda (x) ((f f) x)))))) \
             (lambda (f) \
               (lambda (x) \
                 ((lambda (_) x) \
                  (if (null? x) \
                      (f 'x) \
                      x))))) \
        '())" \
     "()"
test "'((lambda (f) \
         ((lambda (t) \
            (set! f t) \
            (f '())) \
          (lambda (x) \
            (if (null? x) \
                (f 'x) \
                x) \
            x))) \
          '())" \
         "()"
test "'((lambda (x y) x) 'a 'b)" "a"
test "'((lambda (x y) y) 'a 'b)" "b"
test "'((lambda (x y z) z) 'a 'b 'c)" "c"
test "'((lambda () 'z))" "z"
test "'((lambda (x y) (x y)) car '(a . b))" "a"
test "'(((lambda (x y) (lambda (p q) y)) 'a 'b) 'c 'd)" "b"
test "'(((lambda (x y) (lambda (p q) x)) 'a 'b) 'c 'd)" "a"
test "'((lambda (x y) ((lambda (x y) (set! x 'z)) 'a 'b) x) 'c 'd)" "c"
test "'((lambda (x y) ((lambda (p q) (set! x 'z)) 'a 'b) x) 'c 'd)" "z"
test "'((lambda x x))" "()"
test "'((lambda x x) 'a)" "(a)"
test "'((lambda x x) 'a 'b)" "(a b)"
test "'((lambda (x . y) x) 'a)" "a"
test "'((lambda (x . y) y) 'a)" "()"
test "'((lambda (x . y) y) 'a 'b)" "(b)"
test "'((lambda (x . y) y) 'a 'b 'c)" "(b c)"
test "'((lambda (x y . z) x) 'a 'b 'c)" "a"
test "'((lambda (x y . z) y) 'a 'b 'c)" "b"
test "'((lambda (x y . z) z) 'a 'b 'c)" "(c)"
test "'(((lambda (g) \
           ((lambda (f) (f f)) \
            (lambda (f) (g (lambda (x) ((f f) x)))))) \
          (lambda (f) \
            (lambda (x) \
              (if (pair? x) \
                  (cons 'i (f (cdr x))) \
                  '())))) \
         '(1 2 3 4 5 6 7 8 9 10))" \
        "(i i i i i i i i i i)"
test "'((lambda (f) \
          ((lambda (t) \
             (set! f t) \
             (f '(1 2 3 4 5 6 7 8 9 10))) \
           (lambda (x) \
             (if (pair? x) \
                 (cons 'i (f (cdr x))) \
                 '())))) \
         '())" \
        "(i i i i i i i i i i)"

echo
exit $err
